@@ -22,6 +22,8 @@ const SCIENTIFIC_NOTATION = /e/i;
 const ZERO_DECIMAL_INTEGER = /\.0+$/;
 const NON_DECIMAL_LITERAL = /^0[box]/;
 const PURE_ANNOTATION_RE = /^\s*[@#]__PURE__\s*$/;
+const HAS_NEWLINE = /[\n\r\u2028\u2029]/;
+const HAS_BlOCK_COMMENT_END = /\*\//;
 
 const { needsParens } = n;
 
@@ -110,6 +112,8 @@ class Printer {
   _endsWithInteger = false;
   _endsWithWord = false;
   _lastCommentLine = 0;
+  _endsWithInnerRaw: boolean = false;
+  _indentInnerComments: boolean = true;
 
   generate(ast: t.Node) {
     this.print(ast);
@@ -149,6 +153,7 @@ class Printer {
     } else {
       this._queue(charCodes.semicolon);
     }
+    this._noLineTerminator = false;
   }
 
   /**
@@ -183,7 +188,9 @@ class Printer {
    * Writes a token that can't be safely parsed without taking whitespace into account.
    */
 
-  word(str: string): void {
+  word(str: string, noLineTerminatorAfter: boolean = false): void {
+    this._maybePrintInnerComments();
+
     // prevent concatenating words and creating // comment out of division and regex
     if (
       this._endsWithWord ||
@@ -196,6 +203,7 @@ class Printer {
     this._append(str, false);
 
     this._endsWithWord = true;
+    this._noLineTerminator = noLineTerminatorAfter;
   }
 
   /**
@@ -220,6 +228,8 @@ class Printer {
    */
 
   token(str: string, maybeNewline = false): void {
+    this._maybePrintInnerComments();
+
     // space is mandatory to avoid outputting <!--
     // http://javascript.spec.whatwg.org/#comment-syntax
     const lastChar = this.getLastChar();
@@ -237,9 +247,12 @@ class Printer {
 
     this._maybeAddAuxComment();
     this._append(str, maybeNewline);
+    this._noLineTerminator = false;
   }
 
   tokenChar(char: number): void {
+    this._maybePrintInnerComments();
+
     // space is mandatory to avoid outputting <!--
     // http://javascript.spec.whatwg.org/#comment-syntax
     const lastChar = this.getLastChar();
@@ -255,6 +268,7 @@ class Printer {
 
     this._maybeAddAuxComment();
     this._appendChar(char);
+    this._noLineTerminator = false;
   }
 
   /**
@@ -526,13 +540,6 @@ class Printer {
     return this._indentRepeat * this._indent;
   }
 
-  ensureNoLineTerminator(fn: () => void) {
-    const { _noLineTerminator } = this;
-    this._noLineTerminator = true;
-    fn();
-    this._noLineTerminator = _noLineTerminator;
-  }
-
   printTerminatorless(node: t.Node, parent: t.Node, isLabel: boolean) {
     /**
      * Set some state that will be modified if a newline has been inserted before any
@@ -550,9 +557,8 @@ class Printer {
      *  `undefined` will be returned and not `foo` due to the terminator.
      */
     if (isLabel) {
-      this.ensureNoLineTerminator(() => {
-        this.print(node, parent);
-      });
+      this._noLineTerminator = true;
+      this.print(node, parent);
     } else {
       const terminatorState = {
         printed: false,
@@ -573,14 +579,16 @@ class Printer {
   print(
     node: t.Node | null,
     parent?: t.Node,
-    noLineTerminator?: boolean,
+    noLineTerminatorAfter?: boolean,
     // trailingCommentsLineOffset also used to check if called from printJoin
-    // it will be ignored if `noLineTerminator||this._noLineTerminator`
+    // it will be ignored if `noLineTerminatorAfter||this._noLineTerminator`
     trailingCommentsLineOffset?: number,
     forceParens?: boolean,
   ) {
     if (!node) return;
 
+    this._endsWithInnerRaw = false;
+
     const nodeType = node.type;
     const format = this.format;
 
@@ -639,21 +647,24 @@ class Printer {
 
     this.exactSource(loc, printMethod.bind(this, node, parent));
 
-    if (noLineTerminator && !this._noLineTerminator) {
+    if (shouldPrintParens) {
+      this._printTrailingComments(node, parent);
+      this.token(")");
+      this._noLineTerminator = noLineTerminatorAfter;
+    } else if (noLineTerminatorAfter && !this._noLineTerminator) {
       this._noLineTerminator = true;
       this._printTrailingComments(node, parent);
-      this._noLineTerminator = false;
     } else {
       this._printTrailingComments(node, parent, trailingCommentsLineOffset);
     }
 
-    if (shouldPrintParens) this.token(")");
-
     // end
     this._printStack.pop();
 
     format.concise = oldConcise;
     this._insideAux = oldInAux;
+
+    this._endsWithInnerRaw = false;
   }
 
   _maybeAddAuxComment(enteredPositionlessNode?: boolean) {
@@ -776,30 +787,62 @@ class Printer {
   }
 
   _printTrailingComments(node: t.Node, parent?: t.Node, lineOffset?: number) {
-    const comments = this._getComments(false, node);
-    if (!comments?.length) return;
-    this._printComments(
-      COMMENT_TYPE.TRAILING,
-      comments,
-      node,
-      parent,
-      lineOffset,
-    );
+    const { innerComments, trailingComments } = node;
+    // We print inner comments here, so that if for some reason they couldn't
+    // be printed in earlier locations they are still printed *somehwere*,
+    // even if at the end of the node.
+    if (innerComments?.length) {
+      this._printComments(
+        COMMENT_TYPE.TRAILING,
+        innerComments,
+        node,
+        parent,
+        lineOffset,
+      );
+    }
+    if (trailingComments?.length) {
+      this._printComments(
+        COMMENT_TYPE.TRAILING,
+        trailingComments,
+        node,
+        parent,
+        lineOffset,
+      );
+    }
   }
 
   _printLeadingComments(node: t.Node, parent: t.Node) {
-    const comments = this._getComments(true, node);
+    const comments = node.leadingComments;
     if (!comments?.length) return;
     this._printComments(COMMENT_TYPE.LEADING, comments, node, parent);
   }
 
-  printInnerComments(node: t.Node, indent = true) {
-    if (!node.innerComments?.length) return;
+  _maybePrintInnerComments() {
+    if (this._endsWithInnerRaw) this.printInnerComments();
+    this._endsWithInnerRaw = true;
+    this._indentInnerComments = true;
+  }
+
+  printInnerComments() {
+    const node = this._printStack[this._printStack.length - 1];
+    const comments = node.innerComments;
+    if (!comments?.length) return;
+
+    const hasSpace = this.endsWith(charCodes.space);
+    const indent = this._indentInnerComments;
+    const printedCommentsCount = this._printedComments.size;
     if (indent) this.indent();
-    this._printComments(COMMENT_TYPE.INNER, node.innerComments, node);
+    this._printComments(COMMENT_TYPE.INNER, comments, node);
+    if (hasSpace && printedCommentsCount !== this._printedComments.size) {
+      this.space();
+    }
     if (indent) this.dedent();
   }
 
+  noIndentInnerCommentsHere() {
+    this._indentInnerComments = false;
+  }
+
   printSequence(
     nodes: t.Node[],
     parent: t.Node,
@@ -860,22 +903,32 @@ class Printer {
     }
   }
 
-  _getComments(leading: boolean, node: t.Node) {
-    // Note, we use a boolean flag here instead of passing in the attribute name as it is faster
-    // because this is called extremely frequently.
-    return (
-      (node && (leading ? node.leadingComments : node.trailingComments)) || null
-    );
-  }
-
-  _printComment(comment: t.Comment, skipNewLines: COMMENT_SKIP_NEWLINE) {
+  // Returns `true` if the comment cannot be printed in this position due to
+  // line terminators, signaling that the print comments loop can stop and
+  // resume printing comments at the next posisble position. This happens when
+  // printing inner comments, since if we have an inner comment with a multiline
+  // there is at least one inner position where line terminators are allowed.
+  _printComment(
+    comment: t.Comment,
+    skipNewLines: COMMENT_SKIP_NEWLINE,
+  ): boolean {
     // Some plugins (such as flow-strip-types) use this to mark comments as removed using the AST-root 'comments' property,
     // where they can't manually mutate the AST node comment lists.
-    if (comment.ignore) return;
+    if (comment.ignore) return false;
 
-    if (this._printedComments.has(comment)) return;
+    if (this._printedComments.has(comment)) return false;
 
-    if (!this.format.shouldPrintComment(comment.value)) return;
+    const noLineTerminator = this._noLineTerminator;
+
+    if (
+      noLineTerminator &&
+      (HAS_NEWLINE.test(comment.value) ||
+        HAS_BlOCK_COMMENT_END.test(comment.value))
+    ) {
+      return true;
+    }
+
+    if (!this.format.shouldPrintComment(comment.value)) return false;
 
     this._printedComments.add(comment);
 
@@ -924,9 +977,12 @@ class Printer {
 
         val = val.replace(/\n(?!$)/g, `\n${" ".repeat(indentSize)}`);
       }
-    } else if (!this._noLineTerminator) {
+    } else if (!noLineTerminator) {
       val = `//${comment.value}`;
     } else {
+      // It was a single-line comment, so it's guaranteed to not
+      // contain newlines and it can be safely printed as a block
+      // comment.
       val = `/*${comment.value}*/`;
     }
 
@@ -936,13 +992,15 @@ class Printer {
     this.source("start", comment.loc);
     this._append(val, isBlockComment);
 
-    if (!isBlockComment && !this._noLineTerminator) {
+    if (!isBlockComment && !noLineTerminator) {
       this.newline(1, true);
     }
 
     if (printNewLines && skipNewLines !== COMMENT_SKIP_NEWLINE.SKIP_TRAILING) {
       this.newline(1);
     }
+
+    return false;
   }
 
   _printComments(
@@ -952,133 +1010,131 @@ class Printer {
     parent?: t.Node,
     lineOffset: number = 0,
   ) {
-    {
-      const nodeLoc = node.loc;
-      const len = comments.length;
-      let hasLoc = !!nodeLoc;
-      const nodeStartLine = hasLoc ? nodeLoc.start.line : 0;
-      const nodeEndLine = hasLoc ? nodeLoc.end.line : 0;
-      let lastLine = 0;
-      let leadingCommentNewline = 0;
+    const nodeLoc = node.loc;
+    const len = comments.length;
+    let hasLoc = !!nodeLoc;
+    const nodeStartLine = hasLoc ? nodeLoc.start.line : 0;
+    const nodeEndLine = hasLoc ? nodeLoc.end.line : 0;
+    let lastLine = 0;
+    let leadingCommentNewline = 0;
 
-      for (let i = 0; i < len; i++) {
-        const comment = comments[i];
+    const maybeNewline = this._noLineTerminator
+      ? function () {}
+      : this.newline.bind(this);
 
-        if (hasLoc && comment.loc && !this._printedComments.has(comment)) {
-          const commentStartLine = comment.loc.start.line;
-          const commentEndLine = comment.loc.end.line;
-          if (type === COMMENT_TYPE.LEADING) {
-            let offset = 0;
-            if (i === 0) {
-              // Because currently we cannot handle blank lines before leading comments,
-              // we always wrap before and after multi-line comments.
-              if (
-                this._buf.hasContent() &&
-                (comment.type === "CommentLine" ||
-                  commentStartLine != commentEndLine)
-              ) {
-                offset = leadingCommentNewline = 1;
-              }
-            } else {
-              offset = commentStartLine - lastLine;
-            }
-            lastLine = commentEndLine;
+    for (let i = 0; i < len; i++) {
+      const comment = comments[i];
 
-            this.newline(offset);
-            this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL);
-
-            if (i + 1 === len) {
-              this.newline(
-                Math.max(nodeStartLine - lastLine, leadingCommentNewline),
-              );
-              lastLine = nodeStartLine;
-            }
-          } else if (type === COMMENT_TYPE.INNER) {
-            const offset =
-              commentStartLine - (i === 0 ? nodeStartLine : lastLine);
-            lastLine = commentEndLine;
-
-            this.newline(offset);
-            this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL);
-
-            if (i + 1 === len) {
-              this.newline(Math.min(1, nodeEndLine - lastLine)); // TODO: Improve here when inner comments processing is stronger
-              lastLine = nodeEndLine;
+      const printed = this._printedComments.has(comment);
+      if (hasLoc && comment.loc && !printed) {
+        const commentStartLine = comment.loc.start.line;
+        const commentEndLine = comment.loc.end.line;
+        if (type === COMMENT_TYPE.LEADING) {
+          let offset = 0;
+          if (i === 0) {
+            // Because currently we cannot handle blank lines before leading comments,
+            // we always wrap before and after multi-line comments.
+            if (
+              this._buf.hasContent() &&
+              (comment.type === "CommentLine" ||
+                commentStartLine != commentEndLine)
+            ) {
+              offset = leadingCommentNewline = 1;
             }
           } else {
-            const offset =
-              commentStartLine -
-              (i === 0 ? nodeEndLine - lineOffset : lastLine);
-            lastLine = commentEndLine;
+            offset = commentStartLine - lastLine;
+          }
+          lastLine = commentEndLine;
 
-            this.newline(offset);
-            this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL);
+          maybeNewline(offset);
+          this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL);
+
+          if (i + 1 === len) {
+            maybeNewline(
+              Math.max(nodeStartLine - lastLine, leadingCommentNewline),
+            );
+            lastLine = nodeStartLine;
+          }
+        } else if (type === COMMENT_TYPE.INNER) {
+          const offset =
+            commentStartLine - (i === 0 ? nodeStartLine : lastLine);
+          lastLine = commentEndLine;
+
+          maybeNewline(offset);
+          if (this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL)) break;
+
+          if (i + 1 === len) {
+            maybeNewline(Math.min(1, nodeEndLine - lastLine)); // TODO: Improve here when inner comments processing is stronger
+            lastLine = nodeEndLine;
           }
         } else {
-          hasLoc = false;
+          const offset =
+            commentStartLine - (i === 0 ? nodeEndLine - lineOffset : lastLine);
+          lastLine = commentEndLine;
 
-          if (len === 1) {
-            const singleLine = comment.loc
-              ? comment.loc.start.line === comment.loc.end.line
-              : !comment.value.includes("\n");
+          maybeNewline(offset);
+          this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL);
+        }
+      } else {
+        hasLoc = false;
 
-            const shouldSkipNewline =
-              singleLine &&
-              !isStatement(node) &&
-              !isClassBody(parent) &&
-              !isTSInterfaceBody(parent);
+        if (printed) continue;
 
-            if (type === COMMENT_TYPE.LEADING) {
-              this._printComment(
-                comment,
-                (shouldSkipNewline && node.type !== "ObjectExpression") ||
-                  (singleLine && isFunction(parent, { body: node }))
-                  ? COMMENT_SKIP_NEWLINE.SKIP_ALL
-                  : COMMENT_SKIP_NEWLINE.DEFAULT,
-              );
-            } else if (shouldSkipNewline && type === COMMENT_TYPE.TRAILING) {
-              this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL);
-            } else {
-              this._printComment(comment, COMMENT_SKIP_NEWLINE.DEFAULT);
-            }
-          } else if (
-            type === COMMENT_TYPE.INNER &&
-            !(node.type === "ObjectExpression" && node.properties.length > 1) &&
-            node.type !== "ClassBody" &&
-            node.type !== "TSInterfaceBody"
-          ) {
-            // class X {
-            //   /*:: a: number*/
-            //   /*:: b: ?string*/
-            // }
+        if (len === 1) {
+          const singleLine = comment.loc
+            ? comment.loc.start.line === comment.loc.end.line
+            : !HAS_NEWLINE.test(comment.value);
 
+          const shouldSkipNewline =
+            singleLine &&
+            !isStatement(node) &&
+            !isClassBody(parent) &&
+            !isTSInterfaceBody(parent);
+
+          if (type === COMMENT_TYPE.LEADING) {
             this._printComment(
               comment,
-              i === 0
-                ? COMMENT_SKIP_NEWLINE.SKIP_LEADING
-                : i === len - 1
-                ? COMMENT_SKIP_NEWLINE.SKIP_TRAILING
+              (shouldSkipNewline && node.type !== "ObjectExpression") ||
+                (singleLine && isFunction(parent, { body: node }))
+                ? COMMENT_SKIP_NEWLINE.SKIP_ALL
                 : COMMENT_SKIP_NEWLINE.DEFAULT,
             );
+          } else if (shouldSkipNewline && type === COMMENT_TYPE.TRAILING) {
+            if (this._printComment(comment, COMMENT_SKIP_NEWLINE.SKIP_ALL)) {
+              break;
+            }
           } else {
             this._printComment(comment, COMMENT_SKIP_NEWLINE.DEFAULT);
           }
+        } else if (
+          type === COMMENT_TYPE.INNER &&
+          !(node.type === "ObjectExpression" && node.properties.length > 1) &&
+          node.type !== "ClassBody" &&
+          node.type !== "TSInterfaceBody"
+        ) {
+          // class X {
+          //   /*:: a: number*/
+          //   /*:: b: ?string*/
+          // }
+
+          const skippedDueToNewline = this._printComment(
+            comment,
+            i === 0
+              ? COMMENT_SKIP_NEWLINE.SKIP_LEADING
+              : i === len - 1
+              ? COMMENT_SKIP_NEWLINE.SKIP_TRAILING
+              : COMMENT_SKIP_NEWLINE.DEFAULT,
+          );
+          if (skippedDueToNewline) break;
+        } else {
+          this._printComment(comment, COMMENT_SKIP_NEWLINE.DEFAULT);
         }
       }
-
-      if (type === COMMENT_TYPE.TRAILING && hasLoc && lastLine) {
-        this._lastCommentLine = lastLine;
-      }
     }
-  }
-  printAssertions(node: Extract<t.Node, { assertions?: t.ImportAttribute[] }>) {
-    this.space();
-    this.printInnerComments(node);
-    this.token("{");
-    this.space();
-    this.printList(node.assertions, node);
-    this.space();
-    this.token("}");
+
+    if (type === COMMENT_TYPE.TRAILING && hasLoc && lastLine) {
+      this._lastCommentLine = lastLine;
+    }
   }
 }
 
